@@ -4,7 +4,7 @@
 //        for Shared-Memory Multiprocessors
 // Contact author: Emery Berger, http://www.cs.utexas.edu/users/emery
 //
-// Copyright (c) 1998, 1999, The University of Texas at Austin.
+// Copyright (c) 1998-2000, The University of Texas at Austin.
 //
 // This library is free software; you can redistribute it and/or modify
 // it under the terms of the GNU Library General Public License as
@@ -20,26 +20,40 @@
 #ifndef _WRAPPER_H_
 #define _WRAPPER_H_
 
+#include "arch-specific.h"
 #include "config.h"
 #include "processheap.h"
 
 class wrapper {
 public:
-  // Return the process heap.
-  // We make this a static to force it to be
-  // initialized properly before its first use.
-  static processHeap * TheAllocator (void) {
-    static processHeap TheRealAllocator;
-    return &TheRealAllocator;
+  wrapper (void) {
+    hoardLockInit(_lock);
   }
 
+  // Return the process heap.  We can't just have a static
+  // processHeap, because on some systems this causes problems (e.g.,
+  // its destructor is registered by atexit(), which in turn calls
+  // malloc()...)  Instantiating the processHeap inside of a buffer of
+  // characters avoids this problem, since the destructor won't be
+  // added to atexit() at init time.
+  processHeap * TheAllocator (void) {
+    static int initialized = 0;
+    if (!initialized) {
+      hoardLock (_lock);
+      if (!initialized) {
+	TheRealAllocator = new (buf) processHeap;
+	initialized = 1;
+      }
+      hoardUnlock (_lock);
+    }
+    return TheRealAllocator;
+  }
+  
+
 private:
-
-  // Disable copying and assignment.
-
-  wrapper (const wrapper&);
-  const wrapper& operator= (const wrapper&);
-
+  static char buf[sizeof(processHeap)];
+  static processHeap * TheRealAllocator;
+  static hoardLockType _lock;
 };
 
 
